package org.jboss.remoting;

import org.jboss.logging.Logger;
import org.jboss.remoting.transport.ClientInvoker;
import org.jboss.remoting.util.TimerUtil;

import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import EDU.oswego.cs.dl.util.concurrent.ConcurrentHashMap;

/**
 * Internal agent class to ping the remote server to keep lease alive.
 *
 * @author <a href="mailto:telrod@e2technologies.net">Tom Elrod</a>
 */
public class LeasePinger
{
   private static Timer timer = new Timer();
   
   private ClientInvoker client = null;
   private long pingPeriod = -1;
   private long defaultPingPeriod = -1;
   private String invokerSessionId = null;
   private Map clients = new ConcurrentHashMap();
   private TimerTask timerTask = null;

   private static final Logger log = Logger.getLogger(LeasePinger.class);
   private static final boolean isTraceEnabled = log.isTraceEnabled();

   public LeasePinger(ClientInvoker remotingClient, String invokerSessionId, long defaultLeasePeriod)
   {
      this.client = remotingClient;
      this.pingPeriod = defaultLeasePeriod;
      this.defaultPingPeriod = defaultLeasePeriod;
      this.invokerSessionId = invokerSessionId;
   }

   public synchronized void startPing()
   {
      if(isTraceEnabled)
      {
         log.trace("Starting lease timer for client invoker with session ID " + invokerSessionId + " with ping period of " + pingPeriod);
      }
      timerTask = new LeaseTimerTask();
      timer.schedule(timerTask, pingPeriod, pingPeriod);
   }

   public synchronized void stopPing()
   {
      if(isTraceEnabled)
      {
         log.trace("Stopping lease timer for client invoker with session ID " + invokerSessionId);
      }
      if (timerTask != null)
      {
         timerTask.cancel();
         timerTask = null;
         try
         {
            // sending null for the request map, will indicate to server that is full disconnect (for client invoker)
            client.invoke(new InvocationRequest(invokerSessionId, null, "$DISCONNECT$", null, null, null));
         }
         catch (Throwable throwable)
         {
            RuntimeException e = new RuntimeException("Error tearing down lease with server.");
            e.initCause(throwable);
            throw e;
         }
      }
   }

   protected synchronized void sendClientPing()
   {
      if (client != null)
      {
         try
         {
            if(isTraceEnabled)
            {
               String clientSessionIds = "";
               if(clients != null)
               {
                  Collection col = clients.values();
                  Iterator itr = col.iterator();
                  while(itr.hasNext())
                  {
                     ClientHolder h = (ClientHolder)itr.next();
                     clientSessionIds = clientSessionIds + h.getSessionId() + "\n";
                  }
               }
               log.trace("Sending ping to server for client invoker with session ID " + invokerSessionId + ".  " +
                         "Currently managing lease for following clients:\n" + clientSessionIds);
            } // end trace

            Map clientsClone = new ConcurrentHashMap();
            clientsClone.putAll(clients);
            Map requestClients = new ConcurrentHashMap();
            requestClients.put(ClientHolder.CLIENT_HOLDER_KEY, clientsClone);
            client.invoke(new InvocationRequest(invokerSessionId, null, "$PING$", requestClients, null, null));
         }
         catch (Throwable throwable)
         {
            log.warn("Error sending lease ping to server for client invoker with session ID " + invokerSessionId);
         }
      }
   }

   public synchronized void addClient(String sessionId, Map configuration, long leasePeriod)
   {
      if (leasePeriod <= 0)
      {
         leasePeriod = defaultPingPeriod;
      }

      if(isTraceEnabled)
      {
         log.trace("Adding new client to lease for client invoker with session ID " + invokerSessionId + " where " +
                   "client session ID is " + sessionId + " and lease period is " + leasePeriod);
      }
      ClientHolder newClient = new ClientHolder(sessionId, configuration, leasePeriod);
      clients.put(sessionId, newClient);

      sendClientPing();

      // if new client lease period is less than the current ping period, need to refresh to new one
      if (leasePeriod < pingPeriod)
      {
         pingPeriod = leasePeriod;
         // don't want to call stopPing() as that will send disconnect for client invoker
         if (timerTask != null)
         {
            timerTask.cancel();
            timerTask = null;
         }
         startPing();
      }
   }

   public synchronized boolean removeClient(String sessionId)
   {
      boolean isLastClientLease = false;

      if(isTraceEnabled)
      {
         log.trace("Removing client with session ID " + sessionId + " from lease for client invoker with session ID " + invokerSessionId);
      }
      ClientHolder holder = (ClientHolder) clients.remove(sessionId);
      if (holder != null)
      {
         // send disconnect for this client
         try
         {
            Map clientMap = new HashMap();
            clientMap.put(ClientHolder.CLIENT_HOLDER_KEY, holder);
            client.invoke(new InvocationRequest(invokerSessionId, null, "$DISCONNECT$", clientMap, null, null));
            if(isTraceEnabled)
            {
               log.trace("Sent out disconnect message to server for lease tied to client session ID " + sessionId);
            }
         }
         catch (Throwable throwable)
         {
            log.warn("Error sending disconnect for client lease where client session ID is " + sessionId);
         }
      }
      else
      {
         log.warn("Tried to remove lease for client with session ID " + sessionId + ", but did not exist for client invoker lease (session ID " + invokerSessionId + ")");
      }

      if (clients.isEmpty())
      {
         isLastClientLease = true;
         if(isTraceEnabled)
         {
            log.trace("There are no more client leases tied to this client invoker's lease (session ID " + invokerSessionId + ")");
         }
      }
      else
      {
         // now need to see if any of the other client holders have a lower lease period than default
         long tempPingPeriod = defaultPingPeriod;
         Collection clientHolders = clients.values();
         Iterator itr = clientHolders.iterator();
         while (itr.hasNext())
         {
            ClientHolder clientHolder = (ClientHolder) itr.next();
            long clientHolderLeasePeriod = clientHolder.getLeasePeriod();
            if (clientHolderLeasePeriod > 0 && clientHolderLeasePeriod < tempPingPeriod)
            {
               tempPingPeriod = clientHolderLeasePeriod;
            }
         }

         // was there a change in lease period?
         if (tempPingPeriod != pingPeriod)
         {
            // need to update to new ping period and reset timer
            pingPeriod = tempPingPeriod;

            if (timerTask != null)
            {
               timerTask.cancel();
               timerTask = null;
            }
            startPing();
         }

      }
      return isLastClientLease;
   }

   private class LeaseTimerTask extends TimerTask
   {

      public void run()
      {
         sendClientPing();
      }
   }
}
