/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting.transport.bisocket;

import java.io.IOException;
import java.io.OutputStream;
import java.net.Socket;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;

import org.jboss.logging.Logger;
import org.jboss.remoting.Client;
import org.jboss.remoting.ConnectionFailedException;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.invocation.InternalInvocation;
import org.jboss.remoting.marshal.Marshaller;
import org.jboss.remoting.marshal.UnMarshaller;
import org.jboss.remoting.transport.BidirectionalClientInvoker;
import org.jboss.remoting.transport.socket.SocketClientInvoker;

/**
 * The bisocket transport, an extension of the socket transport, is designed to allow
 * a callback server to function behind a firewall.  All connections are created by
 * a Socket constructor or factory on the client side connecting to a ServerSocket on
 * the server side.  When a callback client invoker on the server side needs to
 * open a connection to the callback server, it requests a connection by sending a
 * request message over a control connection to the client side.
 * 
 * Because all connections are created in one direction, the bisocket transport is
 * asymmetric, in the sense that client invokers and server invokers behave differently
 * on the client side and on the server side.
 * 
 * 
 * 
 * @author <a href="mailto:ron.sigal@jboss.com">Ron Sigal</a>
 */
public class BisocketClientInvoker
extends SocketClientInvoker
implements BidirectionalClientInvoker
{  
   private static final Logger log = Logger.getLogger(BisocketClientInvoker.class);
   private static Map listenerIdToClientInvokerMap = Collections.synchronizedMap(new HashMap());
   private static Map listenerIdToCallbackClientInvokerMap = Collections.synchronizedMap(new HashMap());
   private static Map listenerIdToSocketsMap = new HashMap();
   private static Timer timer;
 
   protected String listenerId;
   
   private int pingFrequency = Bisocket.PING_FREQUENCY_DEFAULT;;
   private InvokerLocator secondaryLocator;
   private Socket controlSocket;
   private OutputStream controlOutputStream;
   private Object controlLock = new Object();
   private PingTimerTask pingTimerTask;
   private boolean isCallbackInvoker;

   
   /**
    * 
    * FIXME Comment this
    * 
    * @param listenerId
    * @return
    */
   static BisocketClientInvoker getBisocketClientInvoker(String listenerId)
   {
      return (BisocketClientInvoker) listenerIdToClientInvokerMap.get(listenerId);
   }
   
   
   static BisocketClientInvoker getBisocketCallbackClientInvoker(String listenerId)
   {
      return (BisocketClientInvoker) listenerIdToCallbackClientInvokerMap.get(listenerId);
   }
   
   
   static void transferSocket(String listenerId, Socket socket)
   {
      Set sockets = null;
      
      synchronized (listenerIdToSocketsMap)
      {
         sockets = (Set) listenerIdToSocketsMap.get(listenerId);
         if (sockets == null)
         {
            sockets = new HashSet();
            listenerIdToSocketsMap.put(listenerId, sockets);
         }
      }
      
      synchronized (sockets)
      {
         sockets.add(socket);
         sockets.notify();
      }
   }
   
   
   public BisocketClientInvoker(InvokerLocator locator) throws IOException
   {
      this(locator, null);
   }

   
   public BisocketClientInvoker(InvokerLocator locator, Map config) throws IOException
   {
      super(locator, config);
      
      if (config != null)
      {
         listenerId = (String) config.get(Client.LISTENER_ID_KEY);
         if (listenerId != null)
         {
            isCallbackInvoker = true;
            listenerIdToCallbackClientInvokerMap.put(listenerId, this);
            
            synchronized (listenerIdToSocketsMap)
            {
               if (listenerIdToSocketsMap.get(listenerId) == null)
                  listenerIdToSocketsMap.put(listenerId, new HashSet());
            }
            
            log.debug("registered " + listenerId + " -> " + this);
         }

         // look for socketTimeout param
         Object val = config.get(Bisocket.PING_FREQUENCY);
         if (val != null)
         {
            try
            {
               int nVal = Integer.valueOf((String) val).intValue();
               pingFrequency = nVal;
               log.debug("Setting ping frequency to: " + pingFrequency);
            }
            catch (Exception e)
            {
               log.warn("Could not convert " + Bisocket.PING_FREQUENCY +
                     " value of " + val + " to an int value.");
            }
         }
      }
   }
   
   
   public int getPingFrequency()
   {
      return pingFrequency;
   }
   
   
   public void setPingFrequency(int pingFrequency)
   {
      this.pingFrequency = pingFrequency;
   }
   
   
   protected void handleConnect() throws ConnectionFailedException
   {
      super.handleConnect();
      
      // Callback client on server side.
      if (listenerId != null)
      {
         return;
      }
      
      // Client on client side.
      try
      {
         secondaryLocator = getSecondaryLocator();
      }
      catch (Throwable e)
      {
         log.error("Unable to retrieve address/port of secondary server socket", e);
         throw new ConnectionFailedException(e.getMessage());
      }
   }

   
   protected void handleDisconnect()
   {
      super.handleDisconnect();
      if (listenerId != null)
      {
         if (isCallbackInvoker)
            listenerIdToCallbackClientInvokerMap.remove(listenerId);
         else
            listenerIdToClientInvokerMap.remove(listenerId);

         listenerIdToSocketsMap.remove(listenerId);
         if (pingTimerTask != null)
            pingTimerTask.shutDown();
      }
   }
   
   
   protected Object transport(String sessionId, Object invocation, Map metadata,
                              Marshaller marshaller, UnMarshaller unmarshaller)
   throws IOException, ConnectionFailedException, ClassNotFoundException
   {  
      if (invocation instanceof InvocationRequest)
      {
         InvocationRequest ir = (InvocationRequest) invocation;
         Object o = ir.getParameter();
         if (o instanceof InternalInvocation)
         {
            InternalInvocation ii = (InternalInvocation) o;
            if (InternalInvocation.ADDLISTENER.equals(ii.getMethodName()))
            {
               Map requestPayload = ir.getRequestPayload();
               String listenerId = (String) requestPayload.get(Client.LISTENER_ID_KEY);
               listenerIdToClientInvokerMap.put(listenerId, this);
               BisocketServerInvoker callbackServerInvoker;
               callbackServerInvoker = BisocketServerInvoker.getBisocketServerInvoker(listenerId);
               callbackServerInvoker.createControlConnection(listenerId, secondaryLocator);
            }
         }
      }
      
      return super.transport(sessionId, invocation, metadata, marshaller, unmarshaller);
   }
   
   
   protected Socket createSocket(String address, int port) throws IOException
   {
      if (listenerId == null)
         return super.createSocket(address, port);

      Set sockets = null;
      
      synchronized (listenerIdToSocketsMap)
      {
         sockets = (Set) listenerIdToSocketsMap.get(listenerId);
         if (sockets == null)
         {
            sockets = new HashSet();
            listenerIdToSocketsMap.put(listenerId, sockets);
         }
      }
      
      if (controlSocket == null)
      {
         synchronized (sockets)
         {
            if (sockets.isEmpty())
            {
               while (true)
               {
                  try
                  {
                     sockets.wait(1000);
                     break;
                  }
                  catch (InterruptedException ignored)
                  {
                     log.warn("unexpected interrupt");
                  }
               }
            }
            
            if (sockets.isEmpty())
               throw new IOException("Timed out trying to create control socket");
            
            Iterator it = sockets.iterator();
            controlSocket = (Socket) it.next();
            it.remove();
            controlOutputStream = controlSocket.getOutputStream();
            log.debug("got control socket");
            pingTimerTask = new PingTimerTask();
            if (timer == null)
            {
               timer = new Timer();
            }
            timer.schedule(pingTimerTask, pingFrequency, pingFrequency);
         }
      }
      
      synchronized (controlLock)
      {
         controlOutputStream.write(Bisocket.CREATE_ORDINARY_SOCKET);
      }

      synchronized (sockets)
      {
         if (sockets.isEmpty())
         {
            while (true)
            {
               try
               {
                  sockets.wait(1000);
                  break;
               }
               catch (InterruptedException ignored)
               {
                  log.warn("unexpected interrupt");
               }
            }
         }
         
         if (sockets.isEmpty())
            throw new IOException("Timed out trying to create socket");
         
         Iterator it = sockets.iterator();
         Socket socket = (Socket) it.next();
         it.remove();
         log.debug("socket found");
         return socket;
      }
   }
   
   
   void replaceControlSocket(Socket socket) throws IOException
   {
      synchronized (controlLock)
      {
         controlSocket = socket;
         controlOutputStream = controlSocket.getOutputStream();
         log.debug("replaced control socket");
      }
      
      if (pingTimerTask != null)
         pingTimerTask.cancel();
      
      pingTimerTask = new PingTimerTask();
      if (timer == null)
      {
         timer = new Timer();
      }
      timer.schedule(pingTimerTask, pingFrequency, pingFrequency);
   }
   
   
   InvokerLocator getSecondaryLocator() throws Throwable
   {
      InternalInvocation ii = new InternalInvocation(Bisocket.GET_SECONDARY_INVOKER_LOCATOR, null);
      InvocationRequest r = new InvocationRequest(null, null, ii, null, null, null);
      Object o = invoke(r);
      log.debug("secondary locator: " + o);
      secondaryLocator = (InvokerLocator) o;
      return secondaryLocator;
   }
   
   protected Object checkType(Object o, Class c) throws IOException
   {
      if (c.isInstance(o))
         return o;
      
      throw new IOException(o + "is not an instance of " + c);
   }
   
   protected Object checkNull(Object o, String field) throws IOException
   {
      if (o == null)
         throw new IOException(field + " must not be null");
      
      return o;
   }


   public InvokerLocator getCallbackLocator(Map metadata)
   {
      String transport = (String) metadata.get(Client.CALLBACK_SERVER_PROTOCOL);
      String host = (String) metadata.get(Client.CALLBACK_SERVER_HOST);
      String sPort = (String) metadata.get(Client.CALLBACK_SERVER_PORT);
      int port = -1;
      if (sPort != null)
      {
         try
         {
            port = Integer.parseInt(sPort);
         }
         catch (NumberFormatException e)
         {
            throw new RuntimeException("Can not set internal callback server port as configuration value (" + sPort + " is not a number.");
         }
      }

      return new InvokerLocator(transport, host, port, "callback", metadata);
   }
   
   
   class PingTimerTask extends TimerTask
   {
      public void shutDown()
      {
         cancel();
      }
      
      public void run()
      {
         synchronized (controlLock)
         {
            try
            {
               controlOutputStream.write(Bisocket.PING);
            }
            catch (IOException e)
            {
               log.warn("Unable to send ping: shutting down PingTimerTask");
               pingTimerTask = null;
               cancel();
            }
         }
      }
   }
}