/*
* JBoss, Home of Professional Open Source
* Copyright 2005, JBoss Inc., and individual contributors as indicated
* by the @authors tag. See the copyright.txt in the distribution for a
* full listing of individual contributors.
*
* This is free software; you can redistribute it and/or modify it
* under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2.1 of
* the License, or (at your option) any later version.
*
* This software is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this software; if not, write to the Free
* Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
* 02110-1301 USA, or see the FSF site: http://www.fsf.org.
*/

package org.jboss.remoting.transport.bisocket;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import org.jboss.remoting.Client;
import org.jboss.remoting.InvocationRequest;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.ServerInvocationHandler;
import org.jboss.remoting.invocation.InternalInvocation;
import org.jboss.remoting.transport.PortUtil;
import org.jboss.remoting.transport.socket.LRUPool;
import org.jboss.remoting.transport.socket.ServerThread;
import org.jboss.remoting.transport.socket.SocketServerInvoker;
import org.jboss.logging.Logger;


/**
 *  
 * @author <a href="ron.sigal@jboss.com">Ron Sigal</a>
 * @version $Revision: 1.1.2.5 $
 * <p>
 * Copyright Nov 23, 2006
 * </p>
 */
public class BisocketServerInvoker extends SocketServerInvoker
{
   private static final Logger log = Logger.getLogger(BisocketServerInvoker.class);

   private static Map listenerIdToServerInvokerMap = Collections.synchronizedMap(new HashMap());
   private static Timer timer;

   private Map listenerIdToInvokerLocatorMap = new HashMap();
   private ServerSocket secondaryServerSocket;
   private InvokerLocator secondaryLocator;
   private SecondaryServerSocketThread secondaryServerSocketThread;
   private Map controlConnectionThreadMap = new HashMap();
   private int pingFrequency = Bisocket.PING_FREQUENCY_DEFAULT;
   private int pingWindow = 2 * pingFrequency;
   private ControlMonitorTimerTask controlMonitorTimerTask;
   protected boolean isCallbackServer = false;


   public static BisocketServerInvoker getBisocketServerInvoker(String listenerId)
   {
      return (BisocketServerInvoker) listenerIdToServerInvokerMap.get(listenerId);
   }


   public BisocketServerInvoker(InvokerLocator locator)
   {
      super(locator);
   }


   public BisocketServerInvoker(InvokerLocator locator, Map configuration)
   {
      super(locator, configuration);
   }


   public void start() throws IOException
   {
      if (isCallbackServer)
      {
         if(maxPoolSize <= 0)
         {
            maxPoolSize = MAX_POOL_SIZE_DEFAULT;
         }
         clientpool = new LRUPool(2, maxPoolSize);
         clientpool.create();
         threadpool = new LinkedList();
         running = true;
         started = true;
      }
      else
      {
         super.start();
         InetAddress host = getServerSocket().getInetAddress();
         int freePort = PortUtil.findFreePort(host.getHostAddress());
         if (serverSocketFactory != null)
         {
            secondaryServerSocket = serverSocketFactory.createServerSocket(freePort, 0, host);
         }
         else
         {
            secondaryServerSocket = new ServerSocket(freePort, 0, host);
         }
         secondaryLocator = new InvokerLocator(null, host.getHostAddress(), freePort, null, null);
         secondaryServerSocketThread = new SecondaryServerSocketThread(secondaryServerSocket);
         secondaryServerSocketThread.setName("secondaryServerSocketThread");
         secondaryServerSocketThread.setDaemon(true);
         secondaryServerSocketThread.start();
         log.debug("started secondary port: " + host + ":" + freePort);
      }
   }


   public boolean isTransportBiDirectional()
   {
      return true;
   }


   public void createControlConnection(String listenerId, InvokerLocator locator)
   throws IOException
   {
      boolean firstConnection;

      // restarting connection
      if (locator == null)
      {
         firstConnection = false;
         BisocketClientInvoker clientInvoker = BisocketClientInvoker.getBisocketClientInvoker(listenerId);

         try
         {
            locator = clientInvoker.getSecondaryLocator();
         }
         catch (Throwable t)
         {
            log.error("unable to get secondary locator");
            throw new IOException("unable to get secondary locator: " + t.getMessage());
         }
      }
      // first connection
      else
      {
         firstConnection = true;
         listenerIdToInvokerLocatorMap.put(listenerId, locator);
      }

      log.debug("creating control connection: " + locator);

      Socket socket = null;
      if (socketFactory != null)
         socket = socketFactory.createSocket(locator.getHost(), locator.getPort());
      else
         socket = new Socket(locator.getHost(), locator.getPort());

      DataOutputStream dos = new DataOutputStream(socket.getOutputStream());
      if (firstConnection)
      {
         dos.write(Bisocket.CREATE_ORDINARY_SOCKET);
      }
      else
      {
         dos.write(Bisocket.CREATE_CONTROL_SOCKET);
      }
      dos.writeUTF(listenerId);
      Thread thread = new ControlConnectionThread(socket, listenerId);
      thread.setName("control: " + socket.toString());
      thread.setDaemon(true);

      synchronized (controlConnectionThreadMap)
      {
         controlConnectionThreadMap.put(listenerId, thread);
      }

      thread.start();
      log.debug("created control connection: " + socket.toString());

      if (controlMonitorTimerTask == null)
      {
         if (timer == null)
         {
            timer = new Timer();
         }
         controlMonitorTimerTask = new ControlMonitorTimerTask();
         timer.schedule(controlMonitorTimerTask, pingFrequency, pingFrequency);
      }
   }


   public int getPingFrequency()
   {
      return pingFrequency;
   }


   public void setPingFrequency(int pingFrequency)
   {
      this.pingFrequency = pingFrequency;
      pingWindow = 2 * pingFrequency;
   }


   protected void setup() throws Exception
   {
      Object o = configuration.get(Bisocket.IS_CALLBACK_SERVER);

      if (o != null)
      {
         if (o instanceof String)
            isCallbackServer = Boolean.valueOf((String) o).booleanValue();
         else if (o instanceof Boolean)
            isCallbackServer = ((Boolean) o).booleanValue();
         else
            log.error("unrecognized value for configuration key \"" +
                  Bisocket.IS_CALLBACK_SERVER + "\": " + o);
      }

      super.setup();
   }


   protected void cleanup()
   {
      super.cleanup();

      if (controlMonitorTimerTask != null)
         controlMonitorTimerTask.shutdown();

      Iterator it = controlConnectionThreadMap.values().iterator();
      while (it.hasNext())
      {
         ControlConnectionThread t = (ControlConnectionThread) it.next();
         t.shutdown();
         it.remove();
      }

      if (secondaryServerSocketThread != null)
         secondaryServerSocketThread.shutdown();

      if (secondaryServerSocket != null)
      {
         try
         {
            secondaryServerSocket.close();
         }
         catch (IOException e)
         {
            log.info("Error closing secondary server socket: " + e.getMessage());
         }
      }
   }


   protected InvokerLocator getSecondaryLocator()
   {
      return secondaryLocator;
   }

   protected ServerSocket getServerSocket()
   {
      return serverSocket;
   }


   protected Object handleInternalInvocation(InternalInvocation ii,
                                             InvocationRequest ir,
                                             ServerInvocationHandler handler)
   throws Throwable
   {
      if(Bisocket.GET_SECONDARY_INVOKER_LOCATOR.equals(ii.getMethodName()))
      {
         return secondaryLocator;
      }

      Object response = super.handleInternalInvocation(ii, ir, handler);

      if(InternalInvocation.ADDCLIENTLISTENER.equals(ii.getMethodName()))
      {
         Map metadata = ir.getRequestPayload();
         if(metadata != null)
         {
            String listenerId = (String) metadata.get(Client.LISTENER_ID_KEY);
            if (listenerId != null)
            {
               listenerIdToServerInvokerMap.put(listenerId, this);
            }
         }
      }

      return response;
   }


   class ControlConnectionThread extends Thread
   {
      private Socket controlSocket;
      private String listenerId;
      private DataInputStream dis;
      private boolean running;
      private int errorCount;
      private long lastPing = -1;

      ControlConnectionThread(Socket socket, String listenerId) throws IOException
      {
         controlSocket = socket;
         this.listenerId = listenerId;
         dis = new DataInputStream(socket.getInputStream());
      }

      void shutdown()
      {
         running = false;
         synchronized (controlConnectionThreadMap)
         {
            controlConnectionThreadMap.remove(this);
         }
         try
         {
            controlSocket.close();
         }
         catch (IOException e)
         {
            log.warn("unable to close controlSocket");
         }
         interrupt();
      }

      boolean checkConnection()
      {
         if (lastPing < 0)
         {
            return true;
         }
         long currentTime = System.currentTimeMillis();

         if (log.isTraceEnabled())
         {
            log.trace("elapsed: " + (currentTime - lastPing));
         }
         return (currentTime - lastPing <= pingWindow);
      }

      String getListenerId()
      {
         return listenerId;
      }

      public void run()
      {
         running = true;
         while (running)
         {
            Socket socket = null;

            try
            {
               int action = dis.read();
               lastPing = System.currentTimeMillis();

               switch (action)
               {
                  case Bisocket.CREATE_ORDINARY_SOCKET:
                     InvokerLocator locator = (InvokerLocator) listenerIdToInvokerLocatorMap.get(listenerId);
                     if (socketFactory != null)
                        socket = socketFactory.createSocket(locator.getHost(), locator.getPort());
                     else
                        socket = new Socket(locator.getHost(), locator.getPort());
                     DataOutputStream dos = new DataOutputStream(socket.getOutputStream());
                     dos.write(Bisocket.CREATE_ORDINARY_SOCKET);
                     dos.writeUTF(listenerId);
                     break;

                  case Bisocket.PING:
                     continue;

                  case -1:
                     shutdown();
                     return;

                  default:
                     log.error("unrecognized action: " + action);
                     continue;
               }
            }
            catch (IOException e)
            {
               if (running)
               {
                  if ("socket closed".equals(e.getMessage()) ||
                      "Socket is closed".equals(e.getMessage()) ||
                      "Connection reset".equals(e.getMessage()))
                  {
                     shutdown();
                     return;
                  }
                  log.error("Unable to read from control connection: " + e.getMessage());
                  e.printStackTrace();
                  if (++errorCount > 5)
                  {
                     shutdown();
                     return;
                  }
                  continue;
               }

               return;
            }

            synchronized (clientpool)
            {
               if(clientpool.size() < maxPoolSize)
               {
                  Thread thread = null;
                  try
                  {
                     thread = new ServerThread(socket, BisocketServerInvoker.this,
                                               clientpool, threadpool,
                                               getTimeout(), serverSocketClass);
                     thread.start();

                     if (log.isDebugEnabled())
                        log.debug("created: " + thread);
                  }
                  catch (Exception e)
                  {
                     log.error("Unable to create new ServerThread: " + e.getMessage());
                     e.printStackTrace();
                  }

                  synchronized (threadpool)
                  {
                     threadpool.add(thread);
                  }
               }
            }
         }
      }
   }


   class SecondaryServerSocketThread extends Thread
   {
      private ServerSocket secondaryServerSocket;
      boolean running = true;

      SecondaryServerSocketThread(ServerSocket secondaryServerSocket) throws IOException
      {
         this.secondaryServerSocket = secondaryServerSocket;
      }

      void shutdown()
      {
         running = false;
         interrupt();
      }

      public void run()
      {
         while (running)
         {
            try
            {
               Socket socket = secondaryServerSocket.accept();
               log.debug("accepted: " + socket);
               DataInputStream dis = new DataInputStream(socket.getInputStream());
               int action = dis.read();
               String listenerId = dis.readUTF();

               switch (action)
               {
                  case Bisocket.CREATE_CONTROL_SOCKET:
                     BisocketClientInvoker invoker;
                     invoker =  BisocketClientInvoker.getBisocketCallbackClientInvoker(listenerId);
                     if (invoker == null)
                     {
                        log.debug("SecondaryServerSocketThread: transferring socket: " + listenerId);
                        BisocketClientInvoker.transferSocket(listenerId, socket);
                     }
                     else
                     {
                        invoker.replaceControlSocket(socket);
                        log.debug("SecondaryServerSocketThread: created secondary socket: " + listenerId);
                     }
                     break;

                  case Bisocket.CREATE_ORDINARY_SOCKET:
                     BisocketClientInvoker.transferSocket(listenerId, socket);
                     log.debug("SecondaryServerSocketThread: transferred socket: " + listenerId);
                     break;

                  default:
                     log.error("unrecognized action: " + action);
               }
            }
            catch (IOException e)
            {
               if (running)
                  log.error("Failed to accept socket connection", e);
               else
                  return;

            }
         }
      }

      ServerSocket getServerSocket()
      {
         return secondaryServerSocket;
      }
   }


   class ControlMonitorTimerTask extends TimerTask
   {
      private boolean running = true;

      void shutdown()
      {
         running = false;
         cancel();
      }

      public void run()
      {
         if (log.isTraceEnabled())
            log.trace("checking connections");

         Collection controlConnectionThreads = null;
         synchronized (controlConnectionThreadMap)
         {
            controlConnectionThreads = new HashSet(controlConnectionThreadMap.values());
         }

         if (controlConnectionThreads.isEmpty())
            cancel();

         Iterator it = controlConnectionThreads.iterator();
         while (it.hasNext() & running)
         {
            final ControlConnectionThread t = (ControlConnectionThread) it.next();
            if (!t.checkConnection())
            {
               log.warn(this + ": detected failure on control connection " + t + ": requesting new control connection");
               t.shutdown();

               synchronized (controlConnectionThreadMap)
               {
                  controlConnectionThreadMap.remove(t.getListenerId());
               }

               if (!running)
                  return;

               new Thread()
               {
                  public void run()
                  {
                     try
                     {
                        createControlConnection(t.getListenerId(), null);
                     }
                     catch (IOException e)
                     {
                        InvokerLocator locator = (InvokerLocator) listenerIdToInvokerLocatorMap.get(t.getListenerId());
                        log.error("Unable to recreate control connection: " + locator, e);
                     }
                  }
               }.start();

            }
         }
      }
   }
}