package org.jboss.remoting.transport.socket;

import org.jboss.logging.Logger;
import org.jboss.remoting.CannotConnectException;
import org.jboss.remoting.ConnectionFailedException;
import org.jboss.remoting.InvokerLocator;
import org.jboss.remoting.RemoteClientInvoker;
import org.jboss.remoting.Version;
import org.jboss.remoting.serialization.ClassLoaderUtility;
import org.jboss.remoting.marshal.Marshaller;
import org.jboss.remoting.marshal.UnMarshaller;
import org.jboss.remoting.marshal.serializable.SerializableMarshaller;
import org.jboss.util.propertyeditor.PropertyEditors;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.lang.reflect.Constructor;
import java.net.InetAddress;
import java.net.Socket;
import java.net.SocketException;
import java.rmi.MarshalException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Properties;

/**
 * SocketClientInvoker uses Sockets to remotely connect to the a remote ServerInvoker, which
 * must be a SocketServerInvoker.
 *
 * @author <a href="mailto:jhaynie@vocalocity.net">Jeff Haynie</a>
 * @author <a href="mailto:telrod@e2technologies.net">Tom Elrod</a>
 * @version $Revision: 1.16.2.6 $
 */
public class MicroSocketClientInvoker extends RemoteClientInvoker
{
   private static final Logger log = Logger.getLogger(MicroSocketClientInvoker.class);
   private static final boolean trace = log.isTraceEnabled();

   protected InetAddress addr;
   protected int port;

   /**
    * can be either true or false and will indicate if client socket should have TCP_NODELAY turned on or off.
    * TCP_NODELAY is for a specific purpose; to disable the Nagle buffering algorithm.
    * It should only be set for applications that send frequent small bursts of information without
    * getting an immediate response; where timely delivery of data is required
    * (the canonical example is mouse movements). The default is false.
    */
   public static final String TCP_NODELAY_FLAG = "enableTcpNoDelay";

   /**
    * the client side maximum number of threads. The default is MAX_POOL_SIZE.
    */
   public static final String MAX_POOL_SIZE_FLAG = "clientMaxPoolSize";

   /**
    * Specifies the fully qualified class name for the custom SocketWrapper implementation
    * to use on the client. Note, will need to make sure this is marked as a client
    * parameter (using the 'isParam' attribute). Making this change will not
    * affect the marshaller/unmarshaller that is used, which may also be a requirement.
    */
   public static final String CLIENT_SOCKET_CLASS_FLAG = "clientSocketClass";

   /**
    * Default value for enable TCP nodelay.  Value is false.
    */
   public static final boolean TCP_NODELAY_DEFAULT = false;

   /**
    * Indicates if will check the socket connection when
    * getting from pool by sending byte over the connection
    * to validate is still good.
    */
   protected boolean shouldCheckConnection = false;

   // Performance measurements
   public static long getSocketTime = 0;
   public static long readTime = 0;
   public static long writeTime = 0;
   public static long serializeTime = 0;
   public static long deserializeTime = 0;

   /**
    * If the TcpNoDelay option should be used on the socket.
    */
   protected boolean enableTcpNoDelay = TCP_NODELAY_DEFAULT;

   protected String clientSocketClassName = ClientSocketWrapper.class.getName();
   private Constructor clientSocketConstructor = null;
   protected Class clientSocketClass = null;

   /**
    * Default max number of retries to get a valid socket from the
    * socket pool.  This also translates to number of seconds will wait
    * for connection to be returned to connection pool before erroring. Default is 30.
    */
   public static final int MAX_RETRIES = 30;
   public long usedPooled = 0;
   public Object usedPoolLock = new Object();

   protected int numberOfRetries = MAX_RETRIES;

   /**
    * Default max number of times a invocation will be made
    * when it gets a SocketException.  Default is 3.
    */
   public static final int MAX_CALL_RETRIES = 3;

   protected int numberOfCallRetries = MAX_CALL_RETRIES;

   /**
    * Pool for this invoker.  This is shared between all
    * instances of proxies attached to a specific invoker
    */
   protected LinkedList pool = null;

   /**
    * connection information
    */
   protected ServerAddress address;

   protected static final Map connectionPools = new HashMap();

   /**
    * Default max number of socket connections allowed at any
    * point in time.  Default is 50.
    */
   public static final int MAX_POOL_SIZE = 50;

   protected int maxPoolSize = MAX_POOL_SIZE;

   private boolean reuseAddress = true;

   // flag being set on true by a disconnect request. If trying to create a connection goes on in a
   // loop and a disconnect request arrives, this flag will be used to sent this information into
   // the connect loop
   private volatile boolean bailOut;

   public MicroSocketClientInvoker(InvokerLocator locator)
   {
      this(locator, null);
   }

   public MicroSocketClientInvoker(InvokerLocator locator, Map configuration)
   {
      super(locator, configuration);
      try
      {
         setup();
      }
      catch (Exception ex)
      {
         log.error("Error setting up socket client invoker.", ex);
         throw new RuntimeException(ex.getMessage());
      }
   }

   protected void setup()
         throws Exception
   {
      this.addr = InetAddress.getByName(locator.getHost());
      this.port = locator.getPort();

      Properties props = new Properties();
      props.putAll(configuration);
      PropertyEditors.mapJavaBeanProperties(this, props, false);

      configureParameters();

      address = createServerAddress();
   }

   protected ServerAddress createServerAddress()
   {
      return new ServerAddress(addr.getHostAddress(), port, enableTcpNoDelay, -1);
   }

   protected void configureParameters()
   {
      Map params = configuration;
      if (params != null)
      {
         // look for enableTcpNoDelay param
         Object val = params.get(TCP_NODELAY_FLAG);
         if (val != null)
         {
            try
            {
               enableTcpNoDelay = Boolean.valueOf((String) val).booleanValue();
               log.debug(this + " setting SocketClientInvoker::enableTcpNoDelay to " + enableTcpNoDelay);
            }
            catch (Exception e)
            {
               log.warn("Could not convert " + TCP_NODELAY_FLAG + " value of " + val + " to a boolean value.");
            }
         }
         // look for maxPoolSize param
         val = params.get(MAX_POOL_SIZE_FLAG);
         if (val != null)
         {
            try
            {
               maxPoolSize = Integer.valueOf((String)val).intValue();
               log.debug(this + " setting SocketClientInvoker::maxPoolSize to: " + maxPoolSize);
            }
            catch (Exception e)
            {
               log.warn("Could not convert " + MAX_POOL_SIZE_FLAG + " value of " + val + " to a int value");
            }
         }
         // look for client socket class name
         val = params.get(CLIENT_SOCKET_CLASS_FLAG);
         if (val != null)
         {
            String value = (String) val;
            if (value.length() > 0)
            {
               clientSocketClassName = value;
               log.debug(this + " setting ClientSocket class name to " + clientSocketClassName);
            }
         }

         val = params.get(SocketServerInvoker.CHECK_CONNECTION_KEY);
         if (val != null)
         {
            String value = (String) val;
            if (value.length() > 0)
            {
               shouldCheckConnection = Boolean.valueOf(value).booleanValue();
            }
         }
      }
   }

   /**
    * Indicates if will check socket connection when returning from
    * pool by sending byte to the server.  Default value will be false.
    */
   public boolean checkingConnection()
   {
      return shouldCheckConnection;
   }

   /**
    * Returns if newly created sockets will have SO_REUSEADDR enabled.
    * Default is for this to be true.
    */
   public boolean getReuseAddress()
   {
      return reuseAddress;
   }

   /**
    * Sets if newly created socket should have SO_REUSEADDR enable.
    * Default is true.
    * @param reuse
    */
   public void setReuseAddress(boolean reuse)
   {
      reuseAddress = reuse;
   }


   public synchronized void disconnect()
   {
      if (trace) { log.trace(this + " disconnecting ..."); }
      bailOut = true;
      super.disconnect();
   }

   protected void finalize() throws Throwable
   {
      disconnect();
      super.finalize();
   }

   protected synchronized void handleConnect()
         throws ConnectionFailedException
   {
      initPool();
   }

   protected synchronized void handleDisconnect()
   {
      clearPools();
   }

   /**
    * Each implementation of the remote client invoker should have
    * a default data type that is uses in the case it is not specified
    * in the invoker locator uri.
    */
   protected String getDefaultDataType()
   {
      return SerializableMarshaller.DATATYPE;
   }

   /**
    * @param sessionId
    * @param invocation
    * @param marshaller
    * @return
    * @throws java.io.IOException
    * @throws org.jboss.remoting.ConnectionFailedException
    *
    */
   protected Object transport(String sessionId, Object invocation, Map metadata,
                              Marshaller marshaller, UnMarshaller unmarshaller)
         throws IOException, ConnectionFailedException, ClassNotFoundException
   {

      Object response = null;
      long start = System.currentTimeMillis();
      SocketWrapper socketWrapper = null;

      int x = 0;
      SocketException sockEx = null;
      for (; x < numberOfCallRetries; x++)
      {
         try
         {
            socketWrapper = getConnection();
         }
         catch (Exception e)
         {
            throw new CannotConnectException("Can not get connection to server. Problem establishing socket connection for locator - " + locator, e);
         }

         long end = System.currentTimeMillis() - start;
         getSocketTime += end;

         try
         {
            int version = Version.getDefaultVersion();
            boolean performVersioning = Version.performVersioning();

            OutputStream outputStream = socketWrapper.getOutputStream();

            if (performVersioning)
            {
               writeVersion(outputStream, version);
            }

            //TODO: -TME so this is messed up as now ties remoting versioning to using a marshaller type
            versionedWrite(outputStream, marshaller, invocation, version);

            end = System.currentTimeMillis() - start;
            writeTime += end;
            start = System.currentTimeMillis();

            // check to see if is one way invocation and return if is
            if(metadata != null)
            {
               Object val = metadata.get(org.jboss.remoting.Client.ONEWAY_FLAG);
               if(val != null && val instanceof String && Boolean.valueOf((String)val).booleanValue())
               {
                  if(trace) { log.trace(this + " sent oneway invocation, so not waiting for response, returning null"); }
                  return null;
               }
            }

            InputStream inputStream = socketWrapper.getInputStream();
            if (performVersioning)
            {
               version = readVersion(inputStream);
               if (version == -1)
               {
                  throw new SocketException("end of file");
               }
            }

            response = versionedRead(inputStream, unmarshaller, version);

            end = System.currentTimeMillis() - start;
            readTime += end;
         }
         catch (SocketException sex)
         {
            log.debug(this + " got SocketException " + sex);

            try
            {
               socketWrapper.close();
               usedPooled--;
            }
            catch (Exception ex)
            {
               if (trace) { log.trace(this + " couldn't successfully close its socketWrapper", ex); }
            }

            /**
             * About to run out of retries and
             * pool may be full of timed out sockets,
             * so want to flush the pool and try with
             * fresh socket as a last effort.
             */
            if (x == (numberOfCallRetries - 2))
            {
               flushConnectionPool();
            }
            sockEx = sex;
            continue;
         }
         catch (Exception ex)
         {
            log.debug(this + " got exception " + ex);

            try
            {
               socketWrapper.close();
               usedPooled--;
            }
            catch (Exception ignored)
            {
            }
            return handleException(ex, socketWrapper);
         }

         // call worked, so no need to retry
         break;
      }

      // need to check if ran out of retries
      if (x >= numberOfCallRetries)
      {
         handleException(sockEx, socketWrapper);
      }

      // Put socket back in pool for reuse
      synchronized (pool)
      {
         if (pool.size() < maxPoolSize)
         {
            pool.add(socketWrapper);
            synchronized(usedPoolLock)
            {
               usedPooled--;
            }
         }
         else
         {
            if (trace)
            {
               log.trace("Pool was already full, will close the connection");
            }
            try
            {
               socketWrapper.close();
            }
            catch (Exception ignored)
            {
            }
         }
      }

      if (trace) { log.trace(this + " received " + response);  }
      return response;
   }

   public void flushConnectionPool()
   {
      synchronized (pool)
      {
         while (pool != null && pool.size() > 0)
         {
            SocketWrapper socketWrapper = (SocketWrapper) pool.removeFirst();
            try
            {
               socketWrapper.close();
            }
            catch (IOException e)
            {

            }
         }
      }
   }

   protected Object handleException(Exception ex, SocketWrapper socketWrapper)
         throws ClassNotFoundException, MarshalException
   {
      log.error(this + " got marshalling exception, exiting ...", ex);
      if (ex instanceof ClassNotFoundException)
      {
         //TODO: -TME Add better exception handling for class not found exception
         log.error("Error loading classes from remote call result.", ex);
         throw (ClassNotFoundException) ex;
      }

      throw new MarshalException("Failed to communicate. Problem during marshalling/unmarshalling.", ex);
   }

   private Object versionedRead(InputStream inputStream, UnMarshaller unmarshaller, int version) throws IOException, ClassNotFoundException
   {
      //TODO: -TME - is switch required?
      switch (version)
      {
         case Version.VERSION_1:
         case Version.VERSION_2:
         {
            if (trace) { log.trace(this + " reading response from input stream"); }
            return unmarshaller.read(inputStream, null);
         }
         default:
         {
            throw new IOException("Can not read data for version " + version + ".  Supported versions: " + Version.VERSION_1 + ", " + Version.VERSION_2);
         }
      }
   }

   private void versionedWrite(OutputStream outputStream, Marshaller marshaller, Object invocation, int version) throws IOException
   {
      //TODO: -TME Should I worry about checking the version here?  Only one way to do it at this point
      switch (version)
      {
         case Version.VERSION_1:
         case Version.VERSION_2:
         {
            if (trace) { log.trace(this + " writing invocation on output stream"); }
            marshaller.write(invocation, outputStream);
            if (trace) { log.trace(this + " done writing invocation on output stream"); }

            return;
         }
         default:
         {
            throw new IOException("Can not write data for version " + version + ".  Supported versions: " + Version.VERSION_1 + ", " + Version.VERSION_2);
         }
      }
   }

   //TODO: -TME Exact same method in ServerThread
   private int readVersion(InputStream inputStream) throws IOException
   {
      if (trace) { log.trace(this + " reading version from input stream"); }
      int version = inputStream.read();
      if (trace) { log.trace(this + " read version " + version + " from input stream"); }
      return version;
   }

   //TODO: -TME Exact same method in ServerThread
   private void writeVersion(OutputStream outputStream, int version) throws IOException
   {
      if (trace) { log.trace(this + " writing version " + version + " on output stream"); }
      outputStream.write(version);
   }

   /**
    * Close all sockets in a specific pool.
    */
   public static void clearPool(ServerAddress sa)
   {
      try
      {
         LinkedList thepool = (LinkedList) connectionPools.get(sa);
         if (thepool == null)
         {
            return;
         }
         synchronized (thepool)
         {
            int size = thepool.size();
            for (int i = 0; i < size; i++)
            {
               SocketWrapper socketWrapper = (SocketWrapper) thepool.removeFirst();
               try
               {
                  socketWrapper.close();
                  socketWrapper = null;
               }
               catch (Exception ignored)
               {
               }
            }
         }
      }
      catch (Exception ex)
      {
         // ignored
      }
   }

   /**
    * Close all sockets in all pools
    */
   public static void clearPools()
   {
      synchronized (connectionPools)
      {
         Iterator it = connectionPools.keySet().iterator();
         while (it.hasNext())
         {
            ServerAddress sa = (ServerAddress) it.next();

            if (trace) { log.trace("clearing pool for " + sa); }

            clearPool(sa);
         }
      }
   }

   protected void initPool()
   {
      synchronized (connectionPools)
      {
         pool = (LinkedList) connectionPools.get(address);
         if (pool == null)
         {
            pool = new LinkedList();
            connectionPools.put(address, pool);
         }
      }
   }

   /**
    * Sets the number of times an invocation will retry based on getting
    * SocketException.
    *
    * @param numberOfCallRetries
    */
   public void setNumberOfCallRetries(int numberOfCallRetries)
   {
      if (numberOfCallRetries < 1)
      {
         this.numberOfCallRetries = MAX_CALL_RETRIES;
      }
      else
      {
         this.numberOfCallRetries = numberOfCallRetries;
      }
   }

   public int getNumberOfCallRetries()
   {
      return numberOfCallRetries;
   }

   /**
    * Sets the number of retries to get a socket connection.
    *
    * @param numberOfRetries Must be a number greater than 0
    */
   public void setNumberOfRetries(int numberOfRetries)
   {
      if (numberOfRetries < 1)
      {
         this.numberOfRetries = MAX_RETRIES;
      }
      else
      {
         this.numberOfRetries = numberOfRetries;
      }
   }

   public int getNumberOfRetries()
   {
      return numberOfRetries;
   }

   /**
    * used for debugging (tracing) connections leaks
    */
   static int counter = 0;

   protected SocketWrapper getConnection() throws Exception
   {
      SocketWrapper pooled = null;

      //
      // Need to retry a few times
      // on socket connection because, at least on Windoze,
      // if too many concurrent threads try to connect
      // at same time, you get ConnectionRefused
      //
      // Retrying seems to be the most performant.
      //
      // This problem always happens with RMI and seems to
      // have nothing to do with backlog or number of threads
      // waiting in accept() on the server.
      //
      for (int i = 0; i < numberOfRetries; i++)
      {
         synchronized (pool)
         {
            // if connection within pool, use it
            if (pool.size() > 0)
            {
               pooled = getPooledConnection();
            }
         }

         synchronized(usedPoolLock)
         {
            if (pooled != null)
            {
               usedPooled++;
               break;
            }
            else if (usedPooled < maxPoolSize)
            {
               // if no connection in pool and all pooled connections
               // not in use, then need create a new connection which
               // will be latered returned to the pool (thus filling out
               // the pool, since starts out empty).

               if (bailOut)
               {
                  log.debug(this + " has been concurrently disconnected, " +
                     "bailing out from trying to create a new connection");
                  break;
               }

               Socket socket = null;
               long timestamp = System.currentTimeMillis();
               try
               {
                  if (trace)
                  {
                     log.trace(this + " creating socket " + (counter++) + ", attempt " + (i + 1));
                  }
                  socket = createSocket(address.address, address.port);
               }
               catch (Exception ex)
               {
                  log.debug(this + " got Exception " + ex + ", creation attempt took " + (System.currentTimeMillis() - timestamp) + " ms");

                  if (i + 1 < numberOfRetries)
                  {
                     Thread.sleep(1);
                     continue;
                  }
                  throw ex;
               }
               socket.setTcpNoDelay(address.enableTcpNoDelay);
               socket.setReuseAddress(reuseAddress);
               pooled = createClientSocket(socket, address.timeout, getLocator().getParameters());
               usedPooled++;
               break;
            }
         }
         // waiting 1 second (numberOfRetries along with 1 second wait determines timeout on getting pooled connection)
         Thread.sleep(1000);
      }

      if (pooled == null)
      {
         throw new SocketException("Can not obtain client socket connection from pool.  Have waited " +
                                   numberOfRetries + " seconds for available connection (" + usedPooled + " in use)");
      }
      else
      {
         return pooled;
      }
   }

   protected SocketWrapper createClientSocket(Socket socket, int timeout, Map metadata) throws Exception
   {
      if (clientSocketConstructor == null)
      {
         if(clientSocketClass == null)
         {
            clientSocketClass = ClassLoaderUtility.loadClass(clientSocketClassName, getClass());
         }

         clientSocketConstructor = clientSocketClass.getConstructor(new Class[]{Socket.class});

      }

      SocketWrapper clientSocketWrapper = null;
      clientSocketWrapper = (SocketWrapper) clientSocketConstructor.newInstance(new Object[]{socket});

      return clientSocketWrapper;
   }


   protected Socket createSocket(String address, int port) throws IOException
   {
      return new Socket(address, port);
   }

   protected SocketWrapper getPooledConnection()
   {
      SocketWrapper socketWrapper = null;
      while (pool.size() > 0)
      {
         socketWrapper = (SocketWrapper) pool.removeFirst();
         try
         {
            if (socketWrapper != null)
            {
               if (shouldCheckConnection)
               {
                  socketWrapper.checkConnection();
                  return socketWrapper;
               }
               else
               {
                  return socketWrapper;
               }
            }
         }
         catch (Exception ex)
         {
            if (trace)
            {
               log.trace("Couldn't reuse connection from pool");
            }
            try
            {
               socketWrapper.close();
            }
            catch (Exception ignored)
            {
            }
         }
      }
      return null;
   }


   /**
    * The name of of the server.
    */
   public String getServerHostName() throws Exception
   {
      return address.address;
   }

}
